# frozen_string_literal: true

module Gotta
  module Mod
    require "set"

    class MissingDependencyError < StandardError
      def initialize(source, dep)
        msg = "The mod '#{source}' depends on '#{dep}', but '#{dep}' is not installed in this project. Installed mods: #{Mod.hub.loaded_mods.to_a}"
        super msg
      end
    end

    # Starts a new thread and goes on a blocking
    # operation to constantly pop values from the
    # event queue.
    # If the file changed resides within a directory
    # that is being monitored by the hub, all the
    # mods that monitor that file will be executed.
    # The execution of the mod code is done asynchronously
    # on a separate thread.
    class Hub
      attr_reader :queue, :mods, :loaded_mods
      def initialize(queue)
        # This is the separator to generate the key
        # for the mods hash.
        @key_separator = ":::"
        @queue = queue
        @mutex = Mutex.new
        @mods = {}
        @loaded_mods = Set.new
        @paths = Set.new
        @threads = Set.new
      end

      def start
        loop do
          begin
            event = queue.pop
            keys = get_matching_paths_for(event)
            next unless keys.any?
            run_mods(event, keys)
          rescue StandardError => e
            puts e.full_message
            next
          end
        end
      end

      # Add mods to certain events on a certain path.
      # 'path' is relative to the project folder
      # 'on' is :added or :modified or :removed
      def register_mod(mod_name:, path:, on:, &block)
        key = "#{path}#{@key_separator}#{on}"
        @mutex.synchronize {
          @mods[key] ||= []
          @mods[key] << [block, mod_name]
          @paths << path
        }
      end

      def add_dependency(dep)
        @dependencies ||= Set.new
        @dependencies << dep
      end

      def check_dependencies
        @dependencies.each do |deps|
          source, dep = deps.to_a[0]
          next if @loaded_mods.include?(dep)
          raise MissingDependencyError.new(source, dep)
        end
      end

      private
      # Execute the mod code in a separate thread
      def run_mods(event, keys)
        # puts "Running mods for event #{event.inspect}"
        @mutex.synchronize {
          keys.each do |key|
            # puts "Processing key #{key}"
            next unless @mods[key]
            @threads << @mods[key].map do |blk, mod_name|
              Thread.new do
                # puts "Running mod #{mod_name}"
                blk.call(event)
              end
            end
          end
        }
      end

      # Check if the @mods hash has a key with
      # the file/directory path, a separator and the event type.
      # `event.type` can be :added or :modified or :removed
      def get_matching_paths_for(event)
        # puts "MONITORED? #{event.path}#{@key_separator}#{event.type}"
        @mutex.synchronize {
          @paths.map do |p|
            # puts "#{p} -> #{event.path} (#{File.fnmatch(p, event.path)})"
            next unless File.fnmatch(p, event.path)
            "#{p}#{@key_separator}#{event.type}"
          end.compact
        }
      end
    end
  end
end